# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data models for Agent Skills."""

from __future__ import annotations

import re
from typing import Any
from typing import Optional
import unicodedata

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator

from ..features import FeatureName
from ..features import is_feature_enabled

_KEBAB_NAME_PATTERN = re.compile(r"^[a-z0-9]+(-[a-z0-9]+)*$")
_SNAKE_OR_KEBAB_NAME_PATTERN = re.compile(
    r"^([a-z0-9]+(-[a-z0-9]+)*|[a-z0-9]+(_[a-z0-9]+)*)$"
)


class Frontmatter(BaseModel):
  """L1 skill content: metadata parsed from SKILL.md for skill discovery.

  Attributes:
      name: Skill name in kebab-case or snake_case (required).
      description: What the skill does and when the model should use it
        (required).
      license: License for the skill (optional).
      compatibility: Compatibility information for the skill (optional).
      allowed_tools: A space-delimited list of tools that are pre-approved to
        run (optional, experimental). Accepts both ``allowed_tools`` and the
        YAML-friendly ``allowed-tools`` key. For more details, see
        https://agentskills.io/specification#allowed-tools-field.
      metadata: Key-value pairs for client-specific properties (defaults to
        empty dict). For example, to include additional tools, use the
        ``adk_additional_tools`` key with a list of tools.
  """

  model_config = ConfigDict(
      extra="allow",
      populate_by_name=True,
  )

  name: str
  description: str
  license: Optional[str] = None
  compatibility: Optional[str] = None
  allowed_tools: Optional[str] = Field(
      default=None,
      alias="allowed-tools",
      serialization_alias="allowed-tools",
  )
  metadata: dict[str, Any] = {}

  @field_validator("metadata")
  @classmethod
  def _validate_metadata(cls, v: dict[str, Any]) -> dict[str, Any]:
    if "adk_additional_tools" in v:
      tools = v["adk_additional_tools"]
      if not isinstance(tools, list):
        raise ValueError("adk_additional_tools must be a list of strings")
    return v

  @field_validator("name")
  @classmethod
  def _validate_name(cls, v: str) -> str:
    v = unicodedata.normalize("NFKC", v)
    if len(v) > 64:
      raise ValueError("name must be at most 64 characters")
    if is_feature_enabled(FeatureName.SNAKE_CASE_SKILL_NAME):
      pattern = _SNAKE_OR_KEBAB_NAME_PATTERN
      msg = (
          "name must be lowercase kebab-case (a-z, 0-9, hyphens) or"
          " snake_case (a-z, 0-9, underscores), with no leading, trailing,"
          " or consecutive delimiters. Mixing hyphens and underscores is"
          " not allowed."
      )
    else:
      pattern = _KEBAB_NAME_PATTERN
      msg = (
          "name must be lowercase kebab-case (a-z, 0-9,"
          " hyphens), with no leading, trailing, or"
          " consecutive delimiters"
      )
    if not pattern.match(v):
      raise ValueError(msg)
    return v

  @field_validator("description")
  @classmethod
  def _validate_description(cls, v: str) -> str:
    if not v:
      raise ValueError("description must not be empty")
    if len(v) > 1024:
      raise ValueError("description must be at most 1024 characters")
    return v

  @field_validator("compatibility")
  @classmethod
  def _validate_compatibility(cls, v: Optional[str]) -> Optional[str]:
    if v is not None and len(v) > 500:
      raise ValueError("compatibility must be at most 500 characters")
    return v


class Script(BaseModel):
  """Wrapper for script content."""

  src: str

  def __str__(self) -> str:
    """Returns the string representation of the script content.

    This ensures that any script type can be converted to a string, which is
    useful for including the script in prompts or saving it to the file system.
    """
    return self.src


class Resources(BaseModel):
  """L3 skill content: additional instructions, assets, and scripts.

  Attributes:
      references: Additional markdown files with instructions, workflows, or
        guidance.
      assets: Resource materials like database schemas, API documentation,
        templates, or examples.
      scripts: Executable scripts that can be run via bash.
  """

  references: dict[str, str | bytes] = {}
  assets: dict[str, str | bytes] = {}
  scripts: dict[str, Script] = {}

  def get_reference(self, reference_id: str) -> Optional[str | bytes]:
    """Get content of a reference file.

    Args:
        reference_id: Unique path or name of the reference file.

    Returns:
        Reference content as string or bytes, or None if not found
    """
    return self.references.get(reference_id)

  def get_asset(self, asset_id: str) -> Optional[str | bytes]:
    """Get content of an asset file.

    Args:
        asset_id: Unique path or name of the asset file.

    Returns:
        Asset content as string or bytes, or None if not found
    """
    return self.assets.get(asset_id)

  def get_script(self, script_id: str) -> Optional[Script]:
    """Get content of a script file.

    Args:
        script_id: Unique path or name of the script file.

    Returns:
        Script object, or None if not found
    """
    return self.scripts.get(script_id)

  def list_references(self) -> list[str]:
    """List all available reference paths."""
    return list(self.references.keys())

  def list_assets(self) -> list[str]:
    """List all available asset paths."""
    return list(self.assets.keys())

  def list_scripts(self) -> list[str]:
    """List all available script paths."""
    return list(self.scripts.keys())


class Skill(BaseModel):
  """Complete skill representation including frontmatter, instructions, and resources.

  A skill combines:
  - L1: Frontmatter for discovery (name, description).
  - L2: Instructions from SKILL.md body, loaded when skill is triggered.
  - L3: Resources including additional instructions, assets, and scripts,
  loaded as needed.

  Attributes:
      frontmatter: Parsed skill frontmatter from SKILL.md.
      instructions: L2 skill content: markdown instruction from SKILL.md body.
      resources: L3 skill content: additional instructions, assets, and scripts.
  """

  frontmatter: Frontmatter
  instructions: str
  resources: Resources = Resources()

  @property
  def name(self) -> str:
    """Convenience property to access skill name."""
    return self.frontmatter.name

  @property
  def description(self) -> str:
    """Convenience property to access skill description."""
    return self.frontmatter.description



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for skill prompt generation."""

from __future__ import annotations

import html
from typing import Any
from typing import List
from typing import Union
import warnings

from . import models


def format_skills_as_xml(
    skills: List[Union[models.Frontmatter, models.Skill]],
) -> str:
  """Formats available skills into a standard XML string.

  Args:
    skills: A list of skill frontmatter or full skill objects.

  Returns:
      XML string with <available_skills> block containing each skill's
      name and description.
  """

  if not skills:
    return "<available_skills>\n</available_skills>"

  lines = ["<available_skills>"]

  for item in skills:
    lines.append("<skill>")
    lines.append("<name>")
    lines.append(html.escape(item.name))
    lines.append("</name>")
    lines.append("<description>")
    lines.append(html.escape(item.description))
    lines.append("</description>")
    lines.append("</skill>")

  lines.append("</available_skills>")

  return "\n".join(lines)


def __getattr__(name: str) -> Any:
  if name == "DEFAULT_SKILL_SYSTEM_INSTRUCTION":

    from ..tools import skill_toolset

    warnings.warn(
        (
            "Importing DEFAULT_SKILL_SYSTEM_INSTRUCTION from"
            " google.adk.skills.prompt is deprecated."
            " Please import it from google.adk.tools.skill_toolset instead."
        ),
        DeprecationWarning,
        stacklevel=2,
    )
    return skill_toolset.DEFAULT_SKILL_SYSTEM_INSTRUCTION
  raise AttributeError(f"module {__name__} has no attribute {name}")



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OAuth2 credential exchanger implementation."""

from __future__ import annotations

import logging
from typing import Optional

from fastapi.openapi.models import OAuth2
from google.adk.auth.auth_credential import AuthCredential
from google.adk.auth.auth_schemes import AuthScheme
from google.adk.auth.auth_schemes import OAuthGrantType
from google.adk.auth.auth_schemes import OpenIdConnectWithConfig
from google.adk.auth.oauth2_credential_util import create_oauth2_session
from google.adk.auth.oauth2_credential_util import update_credential_with_tokens
from google.adk.utils.feature_decorator import experimental
from typing_extensions import override

from .base_credential_exchanger import BaseCredentialExchanger
from .base_credential_exchanger import CredentialExchangeError
from .base_credential_exchanger import ExchangeResult

try:
  from authlib.integrations.requests_client import OAuth2Session

  AUTHLIB_AVAILABLE = True
except ImportError:
  AUTHLIB_AVAILABLE = False

logger = logging.getLogger("google_adk." + __name__)


@experimental
class OAuth2CredentialExchanger(BaseCredentialExchanger):
  """Exchanges OAuth2 credentials from authorization responses."""

  @override
  async def exchange(
      self,
      auth_credential: AuthCredential,
      auth_scheme: Optional[AuthScheme] = None,
  ) -> ExchangeResult:
    """Exchange OAuth2 credential from authorization response.

    if credential exchange failed, the original credential will be returned.

    Args:
        auth_credential: The OAuth2 credential to exchange.
        auth_scheme: The OAuth2 authentication scheme.

    Returns:
        An ExchangeResult object containing the exchanged credential and a
        boolean indicating whether the credential was exchanged.

    Raises:
        CredentialExchangeError: If auth_scheme is missing.
    """
    if not auth_scheme:
      raise CredentialExchangeError(
          "auth_scheme is required for OAuth2 credential exchange"
      )

    if not AUTHLIB_AVAILABLE:
      # If authlib is not available, we cannot exchange the credential.
      # We return the original credential without exchange.
      # The client using this tool can decide to exchange the credential
      # themselves using other lib.
      logger.warning(
          "authlib is not available, skipping OAuth2 credential exchange."
      )
      return ExchangeResult(auth_credential, False)

    if auth_credential.oauth2 and auth_credential.oauth2.access_token:
      return ExchangeResult(auth_credential, False)

    # Determine grant type from auth_scheme
    grant_type = self._determine_grant_type(auth_scheme)

    if grant_type == OAuthGrantType.CLIENT_CREDENTIALS:
      return await self._exchange_client_credentials(
          auth_credential, auth_scheme
      )
    elif grant_type == OAuthGrantType.AUTHORIZATION_CODE:
      return await self._exchange_authorization_code(
          auth_credential, auth_scheme
      )
    else:
      logger.warning("Unsupported OAuth2 grant type: %s", grant_type)
      return ExchangeResult(auth_credential, False)

  def _determine_grant_type(
      self, auth_scheme: AuthScheme
  ) -> Optional[OAuthGrantType]:
    """Determine the OAuth2 grant type from the auth scheme.

    Args:
        auth_scheme: The OAuth2 authentication scheme.

    Returns:
        The OAuth2 grant type or None if cannot be determined.
    """
    if isinstance(auth_scheme, OAuth2) and auth_scheme.flows:
      return OAuthGrantType.from_flow(auth_scheme.flows)
    elif isinstance(auth_scheme, OpenIdConnectWithConfig):
      # Check supported grant types for OIDC
      if (
          auth_scheme.grant_types_supported
          and "client_credentials" in auth_scheme.grant_types_supported
      ):
        return OAuthGrantType.CLIENT_CREDENTIALS
      else:
        # Default to authorization code if client credentials not supported
        return OAuthGrantType.AUTHORIZATION_CODE

    return None

  async def _exchange_client_credentials(
      self,
      auth_credential: AuthCredential,
      auth_scheme: AuthScheme,
  ) -> ExchangeResult:
    """Exchange client credentials for access token.

    Args:
        auth_credential: The OAuth2 credential to exchange.
        auth_scheme: The OAuth2 authentication scheme.

    Returns:
        An ExchangeResult object containing the exchanged credential and a
        boolean indicating whether the credential was exchanged.
    """
    client, token_endpoint = create_oauth2_session(auth_scheme, auth_credential)
    if not client:
      logger.warning(
          "Could not create OAuth2 session for client credentials exchange"
      )
      return ExchangeResult(auth_credential, False)

    try:
      tokens = client.fetch_token(
          token_endpoint,
          grant_type=OAuthGrantType.CLIENT_CREDENTIALS,
      )
      update_credential_with_tokens(auth_credential, tokens)
      logger.debug("Successfully exchanged client credentials for access token")
    except Exception as e:
      logger.error("Failed to exchange client credentials: %s", e)
      return ExchangeResult(auth_credential, False)

    return ExchangeResult(auth_credential, True)

  def _normalize_auth_uri(self, auth_uri: str | None) -> str | None:
    # Authlib currently used a simplified token check by simply scanning hash
    # existence, yet itself might sometimes add extraneous hashes.
    # Drop trailing empty hash if seen.
    if auth_uri and auth_uri.endswith("#"):
      return auth_uri[:-1]
    return auth_uri

  async def _exchange_authorization_code(
      self,
      auth_credential: AuthCredential,
      auth_scheme: AuthScheme,
  ) -> ExchangeResult:
    """Exchange authorization code for access token.

    Args:
        auth_credential: The OAuth2 credential to exchange.
        auth_scheme: The OAuth2 authentication scheme.

    Returns:
        An ExchangeResult object containing the exchanged credential and a
        boolean indicating whether the credential was exchanged.
    """
    client, token_endpoint = create_oauth2_session(auth_scheme, auth_credential)
    if not client:
      logger.warning(
          "Could not create OAuth2 session for authorization code exchange"
      )
      return ExchangeResult(auth_credential, False)

    try:
      # Authlib already injects client_id for body-based client auth flows such
      # as client_secret_post, so passing it here would duplicate the field.
      tokens = client.fetch_token(
          token_endpoint,
          authorization_response=self._normalize_auth_uri(
              auth_credential.oauth2.auth_response_uri
          ),
          code=auth_credential.oauth2.auth_code,
          grant_type=OAuthGrantType.AUTHORIZATION_CODE,
      )
      update_credential_with_tokens(auth_credential, tokens)
      logger.debug("Successfully exchanged authorization code for access token")
    except Exception as e:
      logger.error("Failed to exchange authorization code: %s", e)
      return ExchangeResult(auth_credential, False)

    return ExchangeResult(auth_credential, True)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from typing import Generic

from ..agents.llm_agent import Agent
from .data_types import AgentWithScoresT
from .data_types import OptimizerResult
from .data_types import SamplingResultT
from .sampler import Sampler


class AgentOptimizer(ABC, Generic[SamplingResultT, AgentWithScoresT]):
  """Base class for agent optimizers."""

  @abstractmethod
  async def optimize(
      self,
      initial_agent: Agent,
      sampler: Sampler[SamplingResultT],
  ) -> OptimizerResult[AgentWithScoresT]:
    """Runs the optimizer.

    Args:
      initial_agent: The initial agent to be optimized.
      sampler: The interface used to get training and validation example UIDs,
        request agent evaluations, and get useful data for optimizing the agent.

    Returns:
      The final result of the optimization process, containing the optimized
      agent instances along with their corresponding scores on the validation
      examples and any optimization metadata.
    """
    ...



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import Any

from google.api_core import exceptions as api_exceptions
from google.auth.credentials import Credentials
from google.cloud import dataplex_v1

from . import client
from .config import BigQueryToolConfig


def _construct_search_query_helper(
    predicate: str, operator: str, items: list[str]
) -> str:
  """Constructs a search query part for a specific predicate and items."""
  if not items:
    return ""

  clauses = [f'{predicate}{operator}"{item}"' for item in items]
  return "(" + " OR ".join(clauses) + ")" if len(items) > 1 else clauses[0]


def search_catalog(
    prompt: str,
    project_id: str,
    *,
    credentials: Credentials,
    settings: BigQueryToolConfig,
    location: str | None = None,
    page_size: int = 10,
    project_ids_filter: list[str] | None = None,
    dataset_ids_filter: list[str] | None = None,
    types_filter: list[str] | None = None,
) -> dict[str, Any]:
  """Finds BigQuery datasets and tables using natural language semantic search via Dataplex.

  Use this tool to discover BigQuery assets when you don't know the exact names.
  It's ideal for searching based on topics, descriptions, or questions about the data.

  Args:
      prompt: The base search query (natural language or keywords).
      project_id: The Google Cloud project ID to scope the search.
      credentials: Credentials for the request.
      settings: BigQuery tool settings.
      location: The Dataplex location to use.
      page_size: Maximum number of results.
      project_ids_filter: Specific project IDs to include in the search results.
        If None, defaults to the scoping project_id.
      dataset_ids_filter: BigQuery dataset IDs to filter by.
      types_filter: Entry types to filter by (e.g., BigQueryEntryType.TABLE,
        BigQueryEntryType.DATASET).

  Returns:
      Search results or error. The "results" list contains items with:
          - name: The Dataplex Entry name (e.g.,
            "projects/p/locations/l/entryGroups/g/entries/e").
          - linked_resource: The underlying BigQuery resource name (e.g.,
            "//bigquery.googleapis.com/projects/p/datasets/d/tables/t").
          - display_name, entry_type, description, location, update_time.

  Examples:
      Search for tables related to customer data:

          >>> search_catalog(
          ...     prompt="Search for tables related to customer data",
          ...     project_id="my-project",
          ...     credentials=creds,
          ...     settings=settings
          ... )
          {
            "status": "SUCCESS",
            "results": [
              {
                "name":
                "projects/my-project/locations/us/entryGroups/@bigquery/entries/entry-id",
                "display_name": "customer_table",
                "entry_type":
                "projects/p/locations/l/entryTypes/bigquery-table",
                "linked_resource":
                "//bigquery.googleapis.com/projects/my-project/datasets/d/tables/customer_table",
                "description": "Table containing customer details.",
                "location": "us",
                "update_time": "2024-01-01 12:00:00+00:00"
              }
            ]
          }
  """

  try:
    if not project_id:
      return {
          "status": "ERROR",
          "error_details": "project_id must be provided.",
      }

    with client.get_dataplex_catalog_client(
        credentials=credentials,
        user_agent=[settings.application_name, "search_catalog"],
    ) as dataplex_client:
      query_parts = []
      if prompt:
        query_parts.append(f"({prompt})")

      # Filter by project IDs
      projects_to_filter = (
          project_ids_filter if project_ids_filter else [project_id]
      )
      if projects_to_filter:
        query_parts.append(
            _construct_search_query_helper("projectid", "=", projects_to_filter)
        )

      # Filter by dataset IDs
      if dataset_ids_filter:
        dataset_resource_filters = []
        for pid in projects_to_filter:
          for did in dataset_ids_filter:
            dataset_resource_filters.append(
                f'linked_resource:"//bigquery.googleapis.com/projects/{pid}/datasets/{did}/*"'
            )
        if dataset_resource_filters:
          query_parts.append(f"({' OR '.join(dataset_resource_filters)})")
      # Filter by entry types
      if types_filter:
        query_parts.append(
            _construct_search_query_helper("type", "=", types_filter)
        )

      # Always scope to BigQuery system
      query_parts.append("system=BIGQUERY")

      full_query = " AND ".join(filter(None, query_parts))

      search_location = location or settings.location or "global"
      search_scope = f"projects/{project_id}/locations/{search_location}"

      request = dataplex_v1.SearchEntriesRequest(
          name=search_scope,
          query=full_query,
          page_size=page_size,
          semantic_search=True,
      )

      response = dataplex_client.search_entries(request=request)

      results = []
      for result in response.results:
        entry = result.dataplex_entry
        source = entry.entry_source
        results.append({
            "name": entry.name,
            "display_name": source.display_name or "",
            "entry_type": entry.entry_type,
            "update_time": str(entry.update_time),
            "linked_resource": source.resource or "",
            "description": source.description or "",
            "location": source.location or "",
        })
      return {"status": "SUCCESS", "results": results}

  except api_exceptions.GoogleAPICallError as e:
    logging.exception("search_catalog tool: API call failed")
    return {"status": "ERROR", "error_details": f"Dataplex API Error: {e}"}
  except Exception as e:
    logging.exception("search_catalog tool: Unexpected error")
    return {"status": "ERROR", "error_details": repr(e)}



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os

from pydantic import alias_generators
from pydantic import BaseModel
from pydantic import ConfigDict

from ...evaluation.eval_case import Invocation
from ...evaluation.evaluation_generator import EvaluationGenerator
from ...evaluation.gcs_eval_set_results_manager import GcsEvalSetResultsManager
from ...evaluation.gcs_eval_sets_manager import GcsEvalSetsManager
from ...sessions.session import Session


class GcsEvalManagers(BaseModel):
  model_config = ConfigDict(
      alias_generator=alias_generators.to_camel,
      populate_by_name=True,
      arbitrary_types_allowed=True,
  )

  eval_sets_manager: GcsEvalSetsManager

  eval_set_results_manager: GcsEvalSetResultsManager


def convert_session_to_eval_invocations(session: Session) -> list[Invocation]:
  """Converts a session data into a list of Invocation.

  Args:
      session: The session that should be converted.

  Returns:
      list: A list of invocation.
  """
  events = session.events if session and session.events else []
  return EvaluationGenerator.convert_events_to_eval_invocations(events)


def create_gcs_eval_managers_from_uri(
    eval_storage_uri: str,
) -> GcsEvalManagers:
  """Creates GcsEvalManagers from eval_storage_uri.

  Args:
      eval_storage_uri: The evals storage URI to use. Supported URIs:
        gs://<bucket name>. If a path is provided, the bucket will be extracted.

  Returns:
      GcsEvalManagers: The GcsEvalManagers object.

  Raises:
      ValueError: If the eval_storage_uri is not supported.
  """
  if eval_storage_uri.startswith('gs://'):
    gcs_bucket = eval_storage_uri.split('://')[1]
    eval_sets_manager = GcsEvalSetsManager(
        bucket_name=gcs_bucket, project=os.environ['GOOGLE_CLOUD_PROJECT']
    )
    eval_set_results_manager = GcsEvalSetResultsManager(
        bucket_name=gcs_bucket, project=os.environ['GOOGLE_CLOUD_PROJECT']
    )
    return GcsEvalManagers(
        eval_sets_manager=eval_sets_manager,
        eval_set_results_manager=eval_set_results_manager,
    )
  else:
    raise ValueError(
        f'Unsupported evals storage URI: {eval_storage_uri}. Supported URIs:'
        ' gs://<bucket name>'
    )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import ClassVar
from typing import Optional

from typing_extensions import override

from ..utils.feature_decorator import experimental
from .eval_case import Invocation
from .eval_metrics import EvalMetric
from .eval_metrics import RubricsBasedCriterion
from .llm_as_judge_utils import get_text_from_content
from .llm_as_judge_utils import get_tool_calls_and_responses_as_json_str
from .llm_as_judge_utils import get_tool_declarations_as_json_str
from .rubric_based_evaluator import RubricBasedEvaluator

logger = logging.getLogger("google_adk." + __name__)

_RUBRIC_BASED_TOOL_USE_QUALITY_V1_PROMPT = """# Mission
- Your mission is to evaluate the quality of responses generated by an AI agent. You will be presented with a user prompt (<user_prompt>), the agent's response (<response>) to that user prompt, and a set of properties (<property>) that you must use to objectively assess the validity of the agent's response.
- Only use the properties provided. Do not make up new properties.
- IMPORTANT: Assess all of the provided properties. Do not drop any of the properties from your response.
- The primary focus of this rating task is to check correctness of the agent's responses w.r.t. each of the properties.

# Rubric
"yes": The agent's response fulfilled the property or the property is not applicable to the response.
"no": The agent's response did not fulfill the property.

# For each property started with a new line, follow these steps:
STEP 1: Repeat the property, word for word, without making any changes. Keep everything including punctuation and capitalization as-is.
STEP 2: Determine the steps needed to **exactly**, **precisely** and **completely** determine whether the agent's response fulfilled the property.
STEP 3: Follow the steps outlined in STEP 2, thinking out loud.
STEP 4: Review the thoughts and the original property.
STEP 5: Output the final verdict.
Property: [[Repeat the property in STEP 1 again.]]
Rationale: [[Explain your reasoning for the verdict.]]
Verdict: [[yes|no]]

# Output format (repeat this format for every property started with a new line):
STEP 1: ...
STEP 2: ...
STEP 3: ...
STEP 4: ...
STEP 5: ...
Property: ...
Rationale: ...
Verdict: ...


# Example output 1

STEP 1: Does the agent run function call 'default_api.grammar_check'?
STEP 2: I need to check if the agent runs the function call with exact function name as 'default_api.grammar_check'.
STEP 3: The response includes a function call 'default_api.grammar_check'.
STEP 4: The function call format and the function name are correct.
STEP 5: yes
Property: Does the agent run function call 'default_api.grammar_check'?
Rationale: The agent's response contains the function call 'default_api.grammar_check' within a proper code block and with the correct function name.
Verdict: yes

STEP 1: Does the agent provide function call 'default_api.grammar_check' with input parameter 'sentence' that is valid compared to the reference 'sentence'= 'the dog walks on the a park' and based on the following guideline? Guideline for 'sentence': 'The wording can differ. The agent response is valid if it conveys similar core content as the reference response. Less efficient and minor inaccurate phrasing is acceptable. The default value is None, if the reference response includes this parameter with value equal to the default value but it is not provided in the agent response, then evaluate it as valid.'
STEP 2: I need to check if the function call 'default_api.grammar_check' includes the parameter 'sentence' and whether the value assigned to 'sentence' is valid according to the provided guideline. The reference value is 'the dog walks on the a park'. According to the guideline, the wording can differ as long as the core content is similar.
STEP 3: The agent's response includes the function call `default_api.grammar_check(sentence="the dog walks on the a park")`. The parameter 'sentence' is present, and the value assigned to it is "the dog walks on the a park", which is identical to the reference value.
STEP 4: The parameter 'sentence' is present and its value is exactly the same as the reference value.
STEP 5: yes
Property: Does the agent provide function call 'default_api.grammar_check' with input parameter 'sentence' that is valid compared to the reference 'sentence'= 'the dog walks on the a park' and based on the following guideline? Guideline for 'sentence': 'The wording can differ. The agent response is valid if it conveys similar core content as the reference response. Less efficient and minor inaccurate phrasing is acceptable. The default value is None, if the reference response includes this parameter with value equal to the default value but it is not provided in the agent response, then evaluate it as valid.'
Rationale: The agent's response includes the 'sentence' parameter in the function call 'default_api.grammar_check', and the value assigned to it is exactly the same as the reference value, thus satisfying the given guideline.
Verdict: yes

# Example output 2

STEP 1: Does the agent run function call 'default_api.search_via_perplexity'?
STEP 2: I need to check if the agent runs the function call with exact function name as 'default_api.search_via_perplexity'.
STEP 3: The response includes a function call `default_api.get_web_search_results`, which does not match 'default_api.search_via_perplexity'.
STEP 4: The function name does not match.
STEP 5: no
Property: Does the agent run function call 'default_api.search_via_perplexity'?
Rationale: The agent called 'default_api.get_web_search_results', not 'default_api.search_via_perplexity'.
Verdict: no

STEP 1: Does the agent provide function call 'default_api.search_via_perplexity' with input parameter 'keyword' that is valid compared to the reference 'keyword'= 'GPT-4o vs GPT-3.5 cost comparison' and based on the following guideline? Guideline for 'keyword': 'The wording can differ. The agent response is valid if it conveys similar core content as the reference response. Less efficient and minor inaccurate phrasing is acceptable.'
STEP 2: Since the previous property is no, this property is not applicable.
STEP 3: N/A
STEP 4: N/A
STEP 5: yes
Property: Does the agent provide function call 'default_api.search_via_perplexity' with input parameter 'keyword' that is valid compared to the reference 'keyword'= 'GPT-4o vs GPT-3.5 cost comparison' and based on the following guideline? Guideline for 'keyword': 'The wording can differ. The agent response is valid if it conveys similar core content as the reference response. Less efficient and minor inaccurate phrasing is acceptable.'
Rationale: The agent did not use the function call 'default_api.search_via_perplexity'.
Verdict: yes


# Available tools, user input, response and properties:
<available_tools>
{tool_declarations}
</available_tools>

<user_prompt>
{user_input}
</user_prompt>

<response>
{tool_usage}
</response>

<properties>
{rubrics}
</properties>

REMEMBER: Your answer will help improve the AI agent. It is important to determine the fulfillment of the properties correctly. Even answering "no" will improve the agent! Respond in pure text, not json.
IMPORTANT: Make sure for each of the property listed, follow the example steps and output "Property: ..." on a new line and "Verdict: ..." on another new line.
"""


@experimental
class RubricBasedToolUseV1Evaluator(RubricBasedEvaluator):
  """An Evaluator for rubric based assessment of the agent's usage of Tools.

  Example: Lets take an example of a Weather Agent that has access to two tools:
  1: GeoCoding Tool: Coverts a city name, address or zip code into geographic
  coordinates.
  2: GetWeather Tool: Gets weather for the next 10 days for the given geographic
  coordinates.

  For this agent, one can create following Rubrics that could focus on tool use

  Rubric 1: A call is made to GeoCoding Tool.
  Rubric 2: A call is made to GetWeather Tool.
  Rubric 3: The call to GetWeather Tool happens after the GeoCoding Tool.
  Rubric 4: The input to GeoCoding Tool can be mapped back to user prompt.
  Rubric 5: The input to GetWeather Tool comes from the output of GeoCoding
  Tool.)

  For each rubric, this evaluator will generate a confidence score between 0
  and 1, where 0 means that agent's response did not satisfy the rubric at all
  and 1 means complete adherence. Value closer to 1 are desirable.

  A combined score using individual rubric confidences will also be generated.
  Like individual rubric confidence scores, the range for this value will be
  between 0 and 1, and it will have the same interpretation.
  """

  criterion_type: ClassVar[type[RubricsBasedCriterion]] = RubricsBasedCriterion
  RUBRIC_TYPE: ClassVar[str] = "TOOL_USE_QUALITY"

  def __init__(self, eval_metric: EvalMetric):
    super().__init__(
        eval_metric,
        criterion_type=RubricBasedToolUseV1Evaluator.criterion_type,
        rubric_type=RubricBasedToolUseV1Evaluator.RUBRIC_TYPE,
    )
    self._auto_rater_prompt_template = _RUBRIC_BASED_TOOL_USE_QUALITY_V1_PROMPT

  @override
  def format_auto_rater_prompt(
      self,
      actual_invocation: Invocation,
      _: Optional[Invocation],
  ) -> str:
    """Returns the autorater prompt."""
    self.create_effective_rubrics_list(actual_invocation.rubrics)
    user_input = get_text_from_content(actual_invocation.user_content)
    tool_usage = get_tool_calls_and_responses_as_json_str(
        actual_invocation.intermediate_data
    )

    rubrics_text = "\n".join([
        f"*  {r.rubric_content.text_property}"
        for r in self._effective_rubrics_list
    ])

    app_details = actual_invocation.app_details
    tool_declarations = "Agent has no tools."
    if app_details:
      tool_declarations = get_tool_declarations_as_json_str(app_details)

    return self._auto_rater_prompt_template.format(
        tool_declarations=tool_declarations,
        user_input=user_input,
        tool_usage=tool_usage,
        rubrics=rubrics_text,
    )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
  from ..memory.memory_entry import MemoryEntry


def extract_text(memory: MemoryEntry, splitter: str = ' ') -> str:
  """Extracts the text from the memory entry."""
  if not memory.content.parts:
    return ''
  return splitter.join(
      [part.text for part in memory.content.parts if part.text]
  )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from google.cloud import speech
from google.genai import types as genai_types

if TYPE_CHECKING:
  from ...agents.invocation_context import InvocationContext


class AudioTranscriber:
  """Transcribes audio using Google Cloud Speech-to-Text."""

  def __init__(self, init_client=False):
    if init_client:
      self.client = speech.SpeechClient()

  def transcribe_file(
      self, invocation_context: InvocationContext
  ) -> list[genai_types.Content]:
    """Transcribe audio, bundling consecutive segments from the same speaker.

    The ordering of speakers will be preserved. Audio blobs will be merged for
    the same speaker as much as we can do reduce the transcription latency.

    Args:
        invocation_context: The invocation context to access the transcription
          cache.

    Returns:
        A list of Content objects containing the transcribed text.
    """

    bundled_audio = []
    current_speaker = None
    current_audio_data = b''
    contents = []

    # Step1: merge audio blobs
    for transcription_entry in invocation_context.transcription_cache or []:
      speaker, audio_data = (
          transcription_entry.role,
          transcription_entry.data,
      )

      if isinstance(audio_data, genai_types.Content):
        if current_speaker is not None:
          bundled_audio.append((current_speaker, current_audio_data))
          current_speaker = None
          current_audio_data = b''
        bundled_audio.append((speaker, audio_data))
        continue

      if not audio_data.data:
        continue

      if speaker == current_speaker:
        current_audio_data += audio_data.data
      else:
        if current_speaker is not None:
          bundled_audio.append((current_speaker, current_audio_data))
        current_speaker = speaker
        current_audio_data = audio_data.data

    # Append the last audio segment if any
    if current_speaker is not None:
      bundled_audio.append((current_speaker, current_audio_data))

    # reset cache
    invocation_context.transcription_cache = []

    # Step2: transcription
    for speaker, data in bundled_audio:
      if isinstance(data, genai_types.Blob):
        audio = speech.RecognitionAudio(content=data)

        config = speech.RecognitionConfig(
            encoding=speech.RecognitionConfig.AudioEncoding.LINEAR16,
            sample_rate_hertz=16000,
            language_code='en-US',
        )

        response = self.client.recognize(config=config, audio=audio)

        for result in response.results:
          transcript = result.alternatives[0].transcript

          parts = [genai_types.Part(text=transcript)]
          role = speaker.lower()
          content = genai_types.Content(role=role, parts=parts)
          contents.append(content)
      else:
        # don't need to transcribe model which are already text
        contents.append(data)

    return contents



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import enum

from .user_simulator_personas import UserBehavior
from .user_simulator_personas import UserPersona
from .user_simulator_personas import UserPersonaRegistry


class PreBuiltBehaviors(enum.Enum):
  """Atomic behaviors that can be mixed and matched to form personas."""

  # --- Advance Behaviors ---
  ADVANCE_DETAIL_ORIENTED = UserBehavior(
      name="Advance in the Agent succeeds",
      description=(
          "The Generated User Response should stick to the Conversation"
          " Plan.When starting a new request, the Generated User Response"
          " should provide all the information required to accomplish a"
          " high-level goal."
      ),
      behavior_instructions=[
          (
              "If the Agent succeeds, make the next request from the"
              " Conversation Plan."
          ),
          "Skip redundant requests already fulfilled by the Agent.",
          (
              "When making a new request, state both the high-level goal you"
              " want to achieve next AND any additional details you need to"
              " achieve that goal."
          ),
      ],
      violation_rubrics=[
          (
              "The Generated User Response repeats a high-level goal that was"
              " already completed in previous turns."
          ),
          (
              "The Generated User Response provides details for a high-level"
              " goal that was already completed."
          ),
          (
              "The Generated User Response response agrees to change the topic"
              " or perform a task not listed in the Conversation Plan."
          ),
          (
              "The Generated User Response invents a new goal not present in"
              " the Conversation Plan."
          ),
          (
              "The Generated User Response invents details (e.g., a made-up"
              " phone number or address) not provided in the Conversation Plan."
          ),
          (
              "The Generated User Response only provides the high-level goal"
              " and the Agent has to ask for additional details."
          ),
          (
              "The Generated User Response tries to accomplish more than one"
              " high-level task in a single turn."
          ),
      ],
  )

  ADVANCE_GOAL_ORIENTED = UserBehavior(
      name="Advance if the Agent succeeds",
      description=(
          "The Generated User Response should stick to the Conversation Plan as"
          " much as possible. It may deviate in response to Agent requests. The"
          " User Simulator starts with high-level goals, expecting the Agent to"
          " ask for specific details."
      ),
      behavior_instructions=[
          (
              "If the Agent succeeds, make the next request from the"
              " Conversation Plan."
          ),
          "Skip redundant requests already fulfilled by the Agent.",
          (
              "When making a request, state only the high-level goal you want"
              " to achieve next."
          ),
          (
              "Do NOT provide any additional information related to the"
              " high-level goal. The Agent must ask for it."
          ),
      ],
      violation_rubrics=[
          (
              "The Generated User Response repeats a high-level goal that was"
              " already completed in previous turns."
          ),
          (
              "The Generated User Response provides details for a high-level"
              " goal that was already completed."
          ),
          (
              "The Generated User Response invents a new goal not present in"
              " the Conversation Plan or in the Agent's messages."
          ),
          (
              "The Generated User Response invents details (e.g., a made-up"
              " phone number or address) not provided in the Conversation Plan"
              " or in the Agent's messages."
          ),
          (
              "The Generated User Response provides specific details for a"
              " high-level goal (email content, recipient address, phone"
              " numbers) BEFORE the Agent has explicitly asked for them."
          ),
          (
              "The Generated User Response tries to accomplish more than one"
              " high-level task in a single turn."
          ),
      ],
  )

  # --- Answering Behaviors ---
  ANSWER_RELEVANT_ONLY = UserBehavior(
      name="Answer only relevant questions",
      description=(
          "The User Simulator should not answer questions that are not relevant"
          ' to the high-level goals in the Conversation Plan (e.g., "How is'
          ' your day going?"). If all questions the Agent asked are not'
          " relevant, the User Simulator should enforce the Conversation Plan"
          ' (e.g., "Please stick to writing the email.").'
      ),
      behavior_instructions=[
          (
              "Only answer the Agent's questions using information from the"
              " Conversation Plan."
          ),
          (
              "Do NOT provide any additional information the Agent did not"
              " explicitly ask for."
          ),
          (
              "If you do not have the information requested by the Agent,"
              " inform the Agent. Do NOT make up information that is not in the"
              " Conversation Plan."
          ),
          (
              "Do NOT answer questions that are not relevant to the high level"
              " goals in the Conversation Plan."
          ),
      ],
      violation_rubrics=[
          "The Agent asked a question that is not relevant to the high-level"
          " goal and the Generated User Response responds to it."
      ],
  )

  ANSWER_ALL = UserBehavior(
      name="Answer all questions",
      description=(
          "The User Simulator should address EVERY question that the Agent"
          ' asked, e.g., if the Agent asks "How is your day going?", the User'
          " Simulator should respond."
      ),
      behavior_instructions=[
          (
              "Only answer the Agent's questions using information from the"
              " Conversation Plan."
          ),
          (
              "Do NOT provide any additional information the Agent did not"
              " explicitly ask for."
          ),
          (
              "If you do not have the information requested by the Agent,"
              " inform the Agent. Do NOT make up information that is not in the"
              " Conversation Plan. Acknowledge you don't know the information."
          ),
      ],
      violation_rubrics=[
          (
              "The Agent asked a question (or multiple questions), and the"
              " Generated User Response failed to address one or all of them."
          ),
          (
              "The Agent asked for information NOT in the Conversation Plan,"
              " and the Generated User Response made up an answer instead of"
              ' stating, e.g., "I don\'t know" or "I don\'t have that info."'
          ),
      ],
  )

  # --- Correcting Behaviors ---
  CORRECT_AGENT = UserBehavior(
      name="Correct the Agent if it makes a mistake",
      description=(
          "The User Simulator should catch and correct the Agent's mistakes."
      ),
      behavior_instructions=[
          "Challenge illogical or incorrect statements made by the Agent.",
          "If the Agent did an incorrect operation, ask the Agent to fix it.",
      ],
      violation_rubrics=[
          (
              "The Agent provided incorrect information, and the Generated User"
              " Response continues as if it was correct."
          ),
          (
              "The Agent made a dangerous assumption (e.g., sending an email"
              " without asking for the content first), and the Generated User"
              " Response continues without correcting the Agent."
          ),
      ],
  )

  DO_NOT_CORRECT_AGENT = UserBehavior(
      name="Do not correct the Agent",
      description=(
          "The User Simulator should end the conversation when the Agent"
          " provides an illogical or incorrect statement."
      ),
      behavior_instructions=[
          (
              "If the Agent made an illogical or incorrect statement, end the"
              " conversation with `{{ stop_signal }}`."
          ),
      ],
      violation_rubrics=[
          "The Agent makes a mistake or an assumption and the Generated User"
          " Response corrects the Agent."
      ],
  )

  # --- Troubleshooting Behaviors ---
  TROUBLESHOOT_ONCE = UserBehavior(
      name="Troubleshoot once (if necessary)",
      description=(
          "The User Simulator should only troubleshoot the Agent ONCE."
          " Troubleshooting is defined as the User Simulator helping the Agent"
          " after the Agent fails to execute an action (e.g., calls a function"
          " incorrectly) or fails to provide a response expected by the"
          " Conversation Plan. Answering a clarification question from the"
          " Agent is NOT troubleshooting. NOTE: Please check the conversation"
          " history count for Agent errors."
      ),
      behavior_instructions=[
          (
              "If the Agent failed to complete a request for the first time,"
              " troubleshoot the failure."
          ),
          (
              "You should only troubleshoot ONCE per conversation. DO NOT"
              " troubleshoot again if the Conversation History shows that the"
              " you have already tried to troubleshoot any request."
          ),
      ],
      violation_rubrics=[
          (
              "The Generated User Response ends the conversation immediately"
              " after the first Agent failure."
          ),
          (
              "On the second Agent failure, the Generated User Response"
              " response continues the conversation without using"
              " `{{ stop_signal }}`."
          ),
          (
              "After the second Agent failure, the Generated User Response"
              " tries to continue the conversation or continues addressing"
              " failures without using `{{ stop_signal }}`."
          ),
      ],
  )

  # --- Ending Behaviors ---
  END_LIMITED_TROUBLESHOOTING = UserBehavior(
      name="End the conversation appropriately",
      description=(
          "A conversation is complete if ANY of the following stop conditions"
          " are true:\n- The Agent has confirmed the completion of all the"
          " high-level goals in the Conversation Plan.\n- The Agent"
          " successfully transferred the User Simulator to a human/live"
          " agent.\n- The Agent failed more than once.\nThe Agent fails if it"
          " is unable to execute an action (e.g., calls a function incorrectly)"
          " or fails to provide a response expected by the Conversation Plan."
          " Asking a clarification question is not a failure."
      ),
      behavior_instructions=[
          (
              "End the conversation only when any of the stopping conditions"
              " are met; do NOT end prematurely."
          ),
          (
              "When ending the conversation because the Agent has completed all"
              " the high-level goals, you must wait until the Agent has"
              " confirmed the completion of all the goals before ending."
          ),
          (
              "Output `{{ stop_signal }}` as part of your response to indicate"
              " that the conversation with the Agent is over."
          ),
          (
              "Pay attention to the Conversation History and count the number"
              " of Agent failures. A second failure should trigger the end of"
              " the conversation."
          ),
      ],
      violation_rubrics=[
          (
              "The conversation meets one of the stop conditions above, but the"
              " Generated User Response did not use `{{ stop_signal }}`."
          ),
          (
              "The Generated User Response used `{{ stop_signal }}` but the"
              " conversation does not meet any of the stop conditions above."
          ),
      ],
  )

  END_NO_TROUBLESHOOTING = UserBehavior(
      name="End the conversation appropriately",
      description=(
          " A conversation is considered completed if ANY of the following stop"
          " conditions are true:\n- The Agent has confirmed the completion of"
          " all the high-level goals in the Conversation Plan.\n- The Agent"
          " successfully transferred the User Simulator to a human/live"
          " agent.\n- The Agent failed.\nThe Agent fails if it is unable to"
          " execute an action (e.g., calls a function incorrectly) or fails to"
          " provide a response expected by the Conversation Plan. Asking a"
          " clarification question is not a failure."
      ),
      behavior_instructions=[
          (
              "End the conversation when any of the stopping conditions are"
              " met; do NOT end prematurely."
          ),
          (
              "When ending the conversation because the Agent has completed all"
              " the high-level goals, you must wait until the Agent has"
              " confirmed the completion of all the goals before ending."
          ),
          (
              "Output `{{ stop_signal }}` as part of your response to indicate"
              " that the conversation with the Agent is over."
          ),
          (
              "Pay attention to the last Agent message in the Conversation"
              " History. If the Agent message contains a failure, end the"
              " conversation."
          ),
      ],
      violation_rubrics=[
          (
              "The conversation meets one of the stop conditions above, but the"
              " Generated User Response did not use `{{ stop_signal }}`."
          ),
          (
              "The Generated User Response used `{{ stop_signal }}` but the"
              " conversation does not meet any of the stop conditions above."
          ),
          (
              "On the first Agent failure, the Generated User Response"
              " continues the conversation without using `{{ stop_signal }}`."
          ),
          (
              "After the first Agent failure, the Generated User Response tries"
              " to continue the conversation without using `{{ stop_signal }}`."
          ),
      ],
  )

  # --- Tone Behaviors ---
  TONE_PROFESSIONAL = UserBehavior(
      name="Professional tone",
      description=(
          "The User Simulator use clear, technical language. NOTE:"
          " `{{ stop_signal }}` is appropriate language."
      ),
      behavior_instructions=[
          "The User Simulator should use clear, technical language.",
          (
              "Avoid slang, frequent abbreviations, emojis, or excessive social"
              " filler and personal asides."
          ),
      ],
      violation_rubrics=[
          (
              'The Generated User Response includes slang (e.g., "gimme,"'
              ' "kinda," "lol"), frequent abbreviations (e.g., "info," "btw"),'
              " or emojis."
          ),
          (
              "The Generated User Response includes significant social filler"
              " or personal asides, e.g., \"Hi there! I hope you're having a"
              " good day."
          ),
          (
              'The Generated User Response is a "wall of text" where a a direct'
              " sentence would suffice."
          ),
          (
              "The tone of the Generated User Response is inconsist with"
              " previous user turns (if present)."
          ),
      ],
  )

  TONE_CONVERSATIONAL = UserBehavior(
      name="Conversational tone",
      description=(
          "The User Simulator sounds informal. NOTE: `{{ stop_signal }}` is"
          " appropriate language."
      ),
      behavior_instructions=[
          (
              "The User Simulator should sound like a normal human having a"
              " casual conversation."
          ),
          (
              "Avoid answers that are too formal in nature or employ overly"
              " polite phrases and expressions."
          ),
          (
              "Avoid answers that lack natural conversational framing, for"
              " example, sterile or purely functional responses."
          ),
      ],
      violation_rubrics=[
          (
              "The Generated User Response is sterile and purely functional"
              " (direct commands) with no natural conversational framing."
          ),
          (
              "The Generated User Response is too formal in nature, employing"
              " overly polite phrases and expressions."
          ),
          (
              'The Generated User Response is a "wall of text" where a simple'
              " sentence would suffice."
          ),
          (
              "The tone of the Generated User Response is inconsist with"
              " previous user turns (if present)."
          ),
      ],
  )


class _PreBuiltPersonas(enum.Enum):
  """A set of pre-defined personas"""

  EXPERT = UserPersona(
      id="EXPERT",
      description=(
          "An Expert knows exactly what they want and views the Agent as a tool"
          " to execute their commands as efficiently as possible. Experts have"
          " little patience for chit-chat or unnecessary questions."
      ),
      behaviors=[
          PreBuiltBehaviors.ADVANCE_DETAIL_ORIENTED.value,
          PreBuiltBehaviors.ANSWER_RELEVANT_ONLY.value,
          PreBuiltBehaviors.CORRECT_AGENT.value,
          PreBuiltBehaviors.TROUBLESHOOT_ONCE.value,
          PreBuiltBehaviors.END_LIMITED_TROUBLESHOOTING.value,
          PreBuiltBehaviors.TONE_PROFESSIONAL.value,
      ],
  )

  NOVICE = UserPersona(
      id="NOVICE",
      description=(
          "A Novice is trying to solve a problem they don't fully understand,"
          " and they rely heavily on the Agent for guidance. Novices are"
          " patient with the Agent's questions, but are unable to troubleshoot"
          " the Agent's mistakes. Novices are also unable to correct the Agent."
      ),
      behaviors=[
          PreBuiltBehaviors.ADVANCE_GOAL_ORIENTED.value,
          PreBuiltBehaviors.DO_NOT_CORRECT_AGENT.value,
          PreBuiltBehaviors.ANSWER_ALL.value,
          PreBuiltBehaviors.END_NO_TROUBLESHOOTING.value,
          PreBuiltBehaviors.TONE_CONVERSATIONAL.value,
      ],
  )

  EVALUATOR = UserPersona(
      id="EVALUATOR",
      description=(
          "An Evaluator is trying to assess whether the Agent can help"
          " accomplish the goals in the Conversation Plan."
      ),
      behaviors=[
          PreBuiltBehaviors.ADVANCE_DETAIL_ORIENTED.value,
          PreBuiltBehaviors.ANSWER_RELEVANT_ONLY.value,
          PreBuiltBehaviors.END_NO_TROUBLESHOOTING.value,
          PreBuiltBehaviors.DO_NOT_CORRECT_AGENT.value,
          PreBuiltBehaviors.TONE_CONVERSATIONAL.value,
      ],
  )


def get_default_persona_registry() -> UserPersonaRegistry:
  registry = UserPersonaRegistry()

  registry.register_persona(
      _PreBuiltPersonas.EXPERT.value.id, _PreBuiltPersonas.EXPERT.value
  )
  registry.register_persona(
      _PreBuiltPersonas.NOVICE.value.id, _PreBuiltPersonas.NOVICE.value
  )
  registry.register_persona(
      _PreBuiltPersonas.EVALUATOR.value.id, _PreBuiltPersonas.EVALUATOR.value
  )

  return registry



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from google.adk.agents.readonly_context import ReadonlyContext
from typing_extensions import override

from . import client
from . import message_tool
from ...features import experimental
from ...features import FeatureName
from ...tools.base_tool import BaseTool
from ...tools.base_toolset import BaseToolset
from ...tools.base_toolset import ToolPredicate
from ...tools.google_tool import GoogleTool
from .config import PubSubToolConfig
from .pubsub_credentials import PubSubCredentialsConfig


@experimental(FeatureName.PUBSUB_TOOLSET)
class PubSubToolset(BaseToolset):
  """Pub/Sub Toolset contains tools for interacting with Pub/Sub topics and subscriptions."""

  def __init__(
      self,
      *,
      tool_filter: ToolPredicate | list[str] | None = None,
      credentials_config: PubSubCredentialsConfig | None = None,
      pubsub_tool_config: PubSubToolConfig | None = None,
  ):
    """Initializes the PubSubToolset.

    Args:
        tool_filter: A predicate or list of tool names to filter the tools in
          the toolset. If None, all tools are included.
        credentials_config: The credentials configuration to use for
          authenticating with Google Cloud.
        pubsub_tool_config: The configuration for the Pub/Sub tools.
    """
    super().__init__(tool_filter=tool_filter)
    self._credentials_config = credentials_config
    self._tool_settings = (
        pubsub_tool_config if pubsub_tool_config else PubSubToolConfig()
    )

  def _is_tool_selected(
      self, tool: BaseTool, readonly_context: ReadonlyContext
  ) -> bool:
    if self.tool_filter is None:
      return True

    if isinstance(self.tool_filter, ToolPredicate):
      return self.tool_filter(tool, readonly_context)

    if isinstance(self.tool_filter, list):
      return tool.name in self.tool_filter

    return False

  @override
  async def get_tools(
      self, readonly_context: ReadonlyContext | None = None
  ) -> list[BaseTool]:
    """Get tools from the toolset."""
    all_tools = [
        GoogleTool(
            func=func,
            credentials_config=self._credentials_config,
            tool_settings=self._tool_settings,
        )
        for func in [
            message_tool.publish_message,
            message_tool.pull_messages,
            message_tool.acknowledge_messages,
        ]
    ]

    return [
        tool
        for tool in all_tools
        if self._is_tool_selected(tool, readonly_context)
    ]

  @override
  async def close(self):
    """Clean up resources used by the toolset."""
    client.cleanup_clients()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handles NL planning related logic."""

from __future__ import annotations

from typing import AsyncGenerator
from typing import Optional
from typing import TYPE_CHECKING

from typing_extensions import override

from ...agents.callback_context import CallbackContext
from ...agents.invocation_context import InvocationContext
from ...agents.readonly_context import ReadonlyContext
from ...events.event import Event
from ...planners.plan_re_act_planner import PlanReActPlanner
from ._base_llm_processor import BaseLlmRequestProcessor
from ._base_llm_processor import BaseLlmResponseProcessor

if TYPE_CHECKING:
  from ...models.llm_request import LlmRequest
  from ...models.llm_response import LlmResponse
  from ...planners.base_planner import BasePlanner


class _NlPlanningRequestProcessor(BaseLlmRequestProcessor):
  """Processor for NL planning."""

  async def run_async(
      self, invocation_context: InvocationContext, llm_request: LlmRequest
  ) -> AsyncGenerator[Event, None]:
    from ...planners.built_in_planner import BuiltInPlanner

    planner = _get_planner(invocation_context)
    if not planner:
      return

    if isinstance(planner, BuiltInPlanner):
      planner.apply_thinking_config(llm_request)
    elif isinstance(planner, PlanReActPlanner):
      if planning_instruction := planner.build_planning_instruction(
          ReadonlyContext(invocation_context), llm_request
      ):
        llm_request.append_instructions([planning_instruction])

      _remove_thought_from_request(llm_request)

    # Maintain async generator behavior
    if False:  # Ensures it behaves as a generator
      yield  # This is a no-op but maintains generator structure


request_processor = _NlPlanningRequestProcessor()


class _NlPlanningResponse(BaseLlmResponseProcessor):

  @override
  async def run_async(
      self, invocation_context: InvocationContext, llm_response: LlmResponse
  ) -> AsyncGenerator[Event, None]:
    from ...planners.built_in_planner import BuiltInPlanner

    if (
        not llm_response
        or not llm_response.content
        or not llm_response.content.parts
    ):
      return

    planner = _get_planner(invocation_context)
    if not planner or isinstance(planner, BuiltInPlanner):
      return

    # Postprocess the LLM response.
    callback_context = CallbackContext(invocation_context)
    processed_parts = planner.process_planning_response(
        callback_context, llm_response.content.parts
    )
    if processed_parts:
      llm_response.content.parts = processed_parts

    if callback_context.state.has_delta():
      state_update_event = Event(
          invocation_id=invocation_context.invocation_id,
          author=invocation_context.agent.name,
          branch=invocation_context.branch,
          actions=callback_context._event_actions,
      )
      yield state_update_event


response_processor = _NlPlanningResponse()


def _get_planner(
    invocation_context: InvocationContext,
) -> Optional[BasePlanner]:
  from ...planners.base_planner import BasePlanner

  agent = invocation_context.agent
  if not hasattr(agent, 'planner'):
    return None
  if not agent.planner:
    return None

  if isinstance(agent.planner, BasePlanner):
    return agent.planner
  return PlanReActPlanner()


def _remove_thought_from_request(llm_request: LlmRequest):
  if not llm_request.contents:
    return

  for content in llm_request.contents:
    if not content.parts:
      continue
    for part in content.parts:
      part.thought = None



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from google.genai import types
from typing_extensions import override

from .base_example_provider import BaseExampleProvider
from .example import Example


class VertexAiExampleStore(BaseExampleProvider):
  """Provides examples from Vertex example store."""

  def __init__(self, examples_store_name: str):
    """Initializes the VertexAiExampleStore.

    Args:
        examples_store_name: The resource name of the vertex example store, in
          the format of
          ``projects/{project}/locations/{location}/exampleStores/{example_store}``.
    """
    self.examples_store_name = examples_store_name

  @override
  def get_examples(self, query: str) -> list[Example]:
    from ..dependencies.vertexai import example_stores

    example_store = example_stores.ExampleStore(self.examples_store_name)
    # Retrieve relevant examples.
    request = {
        "stored_contents_example_parameters": {
            "content_search_key": {
                "contents": [{"role": "user", "parts": [{"text": query}]}],
                "search_key_generation_method": {"last_entry": {}},
            }
        },
        "top_k": 10,
        "example_store": self.examples_store_name,
    }
    response = example_store.api_client.search_examples(request)

    returned_examples = []
    # Convert results to genai formats
    for result in response.results:
      if result.similarity_score < 0.5:
        continue
      expected_contents = [
          content.content
          for content in (
              result.example.stored_contents_example.contents_example.expected_contents
          )
      ]
      expected_output = []
      for content in expected_contents:
        expected_parts = []
        for part in content.parts:
          if part.text:
            expected_parts.append(types.Part.from_text(text=part.text))
          elif part.function_call:
            expected_parts.append(
                types.Part.from_function_call(
                    name=part.function_call.name,
                    args={
                        key: value
                        for key, value in part.function_call.args.items()
                    },
                )
            )
          elif part.function_response:
            expected_parts.append(
                types.Part.from_function_response(
                    name=part.function_response.name,
                    response={
                        key: value
                        for key, value in (
                            part.function_response.response.items()
                        )
                    },
                )
            )
        expected_output.append(
            types.Content(role=content.role, parts=expected_parts)
        )

      returned_examples.append(
          Example(
              input=types.Content(
                  role="user",
                  parts=[
                      types.Part.from_text(
                          text=result.example.stored_contents_example.search_key
                      )
                  ],
              ),
              output=expected_output,
          )
      )
    return returned_examples



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .openapi_spec_parser import OpenApiSpecParser
from .openapi_spec_parser import OperationEndpoint
from .openapi_spec_parser import ParsedOperation
from .openapi_toolset import OpenAPIToolset
from .operation_parser import OperationParser
from .rest_api_tool import AuthPreparationState
from .rest_api_tool import RestApiTool
from .rest_api_tool import snake_to_lower_camel
from .tool_auth_handler import ToolAuthHandler

__all__ = [
    'OpenApiSpecParser',
    'OperationEndpoint',
    'ParsedOperation',
    'OpenAPIToolset',
    'OperationParser',
    'RestApiTool',
    'snake_to_lower_camel',
    'AuthPreparationState',
    'ToolAuthHandler',
]
